#pragma once

#include "alm-params.hpp"
#include "anderson-direction-params.hpp"
#include "anderson-params.hpp"
#include "inner-solve-options.hpp"
#include "lbfgs-direction-params.hpp"
#include "lbfgs-params.hpp"
#include "newton-tr-direction-params.hpp"
#include "panoc-params.hpp"
#include "fista-params.hpp"
#include "pantr-params.hpp"
#include "steihaug-params.hpp"
#include "structured-lbfgs-direction-params.hpp"
#include "structured-newton-direction-params.hpp"
#include "zerofpr-params.hpp"
#if ALPAQA_WITH_LBFGSB
#include "lbfgsb-params.hpp"
#endif
#if ALPAQA_WITH_OCP
#include "panoc-ocp-params.hpp"
#endif
